import logging
import os
from pathlib import Path
import sys

from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context
from oslo_config import cfg

import aprsd
from aprsd_irc_extension import conf  # noqaa
from aprsd_irc_extension.db.models import Base
from aprsd.log import log


home = str(Path.home())
DEFAULT_CONFIG_DIR = f"{home}/.config/aprsd/"
DEFAULT_CONFIG_FILE = f"{home}/.config/aprsd/aprsd.conf"


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")

target_metadata = Base.metadata

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
# fileConfig(config.config_file_name)

folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, folder)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
# target_metadata = haminfo_db.Station.metadata
# target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.
# config_file = context.get_x_argument(as_dictionary=True).get('config_file')
# if not config_file:
print(sys.argv)
config_file = [DEFAULT_CONFIG_FILE]
CONF(
    [], project='aprsd',
    version=aprsd.__version__,
    default_config_files=config_file,
)
log.setup_logging()
# python_logging.captureWarnings(True)
CONF.log_opt_values(LOG, logging.DEBUG)


def get_url():
    url = CONF.aprsd_irc_extension.db_dsn
    assert url, "Couldn't find DB url!!"
    print(f"Using DB URL {url}")
    return url


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    # url = config.get_main_option("sqlalchemy.url")
    url = get_url()
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )
    context.config.set_section_option(config.config_ini_section,
                                      "sqlalchemy.url", url)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    url = get_url()
    # connectable = create_engine(url)
    # context.config.set_main_option('sqlalchemy.url', url)
    context.config.set_section_option(config.config_ini_section,
                                      "sqlalchemy.url", url)
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata,
            compare_type=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import logging
from functools import lru_cache
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from oslo_config import cfg
from aprsd_irc_extension import conf  # noqa

CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


@lru_cache
def get_engine():
    LOG.info(f"Using DB URL {CONF.aprsd_irc_extension.db_dsn}")
    engine = create_engine(CONF.aprsd_irc_extension.db_dsn, pool_pre_ping=True)
    return engine


@lru_cache
def create_session() -> scoped_session:
    engine = get_engine()
    LOG.info(f"Engine {engine}")
    session = scoped_session(
        sessionmaker(autoflush=True, bind=engine)
    )
    return session


class MySession:
    _session = None
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._session = create_session()
        return cls._instance

    @property
    def session(self):
        if not self._session:
            self._session = create_session()
        return self._session

    def close(self):
        self._session.remove()
        self._session = None


def get_session():
    return MySession().session


def close_session():
    MySession().close()
    LOG.debug("session closed")


def init_db_schema(engine):
    from aprsd_irc_extension.db import models
    models.Base.metadata.create_all(engine)
    LOG.info("DB schema initialized")


def wipe_and_init_db_schema(engine):
    from aprsd_irc_extension.db import models
    models.Base.metadata.drop_all(engine)
    models.Base.metadata.create_all(engine)
    LOG.info("DB schema initialized")



import update_checker

import aprsd_irc_extension


def _check_version():
    # check for a newer version
    try:
        check = update_checker.UpdateChecker()
        result = check.check("aprsd-irc-extension", aprsd_irc_extension.__version__)
        if result:
            # Looks like there is an updated version.
            return 1, result
        else:
            return 0, "APRSD IRC Extension is up to date"
    except Exception:
        # probably can't get in touch with pypi for some reason
        # Lets put up an error and move on.  We might not
        # have internet in this aprsd deployment.
        return 1, "Couldn't check for new version of APRSD IRC Extension (aprsd-irc-extension)"



import click

from aprsd.cli_helper import AliasedGroup
from aprsd.main import cli


@cli.group(cls=AliasedGroup, aliases=['irc'], help="IRC related commands for APRSD")
@click.pass_context
def irc(ctx):
    pass



import sys
import unittest


if sys.version_info >= (3, 2):
    from unittest import mock  # noqa
else:
    from unittest import mock  # noqa


class TestMain(unittest.TestCase):

    def test_nothing(self):
        pass



import logging

import click
from oslo_config import cfg

from aprsd import cli_helper
from aprsd.conf import log as aprsd_conf_log

from aprsd_irc_extension.db import session as db_session
import aprsd_irc_extension
from aprsd_irc_extension import cmds
from aprsd_irc_extension import conf  # noqa


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


@cmds.irc.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "-f",
    "--flush",
    "flush",
    is_flag=True,
    show_default=True,
    default=False,
    help="Flush out all old aged messages on disk.",
)
@click.pass_context
@cli_helper.process_standard_options
def db(ctx, flush):
    """Initialize and upgrade the DB schema."""

    LOG.info(f"aprsd-irc-extension version: {aprsd_irc_extension.__version__}")

    CONF.log_opt_values(
        LOG,
        aprsd_conf_log.LOG_LEVELS[CONF.logging.log_level]
    )

    engine = db_session.get_engine()
    db_session.init_db_schema(engine)


@cmds.irc.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "-f",
    "--flush",
    "flush",
    is_flag=True,
    show_default=True,
    default=False,
    help="Flush out all old aged messages on disk.",
)
@click.pass_context
@cli_helper.process_standard_options
def wipe_db(ctx, flush):
    """Completely wipe existing DB and Initialize and upgrade the DB schema."""

    LOG.info(f"aprsd-irc-extension version: {aprsd_irc_extension.__version__}")

    CONF.log_opt_values(
        LOG,
        aprsd_conf_log.LOG_LEVELS[CONF.logging.log_level]
    )

    engine = db_session.get_engine()
    db_session.wipe_and_init_db_schema(engine)



import logging
from typing import List

import sqlalchemy as sa
from sqlalchemy.orm import registry, relationship
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql.schema import ForeignKey, UniqueConstraint
from sqlalchemy.orm import Mapped
import sqlalchemy.ext.declarative as dec

import aprslib
from aprsd_irc_extension.db import session as db_session
from aprsd.packets import core as aprsd_packets

mapper_registry = registry()
ModelBase = dec.declarative_base()

LOG = logging.getLogger("APRSD")


class Base(metaclass=DeclarativeMeta):
    __abstract__ = True

    registry = mapper_registry
    metadata = mapper_registry.metadata

    __init__ = mapper_registry.constructor


class Channel(Base):
    __tablename__ = "channels"
    __allow_unmapped__ = True

    id = sa.Column(sa.Integer, sa.Sequence('channel_id_seq'), primary_key=True)
    name = sa.Column(sa.Text, nullable=False)
    messages: Mapped[List["ChannelMessages"]] = relationship(
        order_by="ChannelMessages.timestamp",
        back_populates="channel"
    )  # type: ignore
    users: Mapped[List["ChannelUsers"]] = relationship(
        order_by="desc(ChannelUsers.user)",
        back_populates="channel",
        cascade="all, delete"
    )  # type: ignore

    def __repr__(self):
        users = [u.user for u in self.users]
        return (f"<Channel(name='{self.name}', users({len(self.users)})='{users}'), "
                f"messages({len(self.messages)})>")

    def to_json(self, include_messages=True):
        our_json = {
            "name": self.name,
            "users": [u.user for u in self.users],
            # "messages": [m.message for m in self.messages]
        }
        if include_messages:
            our_json["messages"] = [m.packet.to_json() for m in self.messages]

        return our_json

    @staticmethod
    def find_by_name(session, name: str) -> "Channel":
        try:
            channel = session.query(
                Channel
            ).filter(
                Channel.name == name
            ).one()
            return channel
        except sa.orm.exc.NoResultFound:
            return None

    @staticmethod
    def create_channel(name: str) -> "Channel":
        session = db_session.get_session()
        channel = Channel(name=name)
        session.add(channel)
        session.commit()
        return channel

    @staticmethod
    def get_all_channels() -> List["Channel"]:
        session = db_session.get_session()
        channels = session.query(Channel).all()
        return channels

    def save(self, session) -> None:
        LOG.debug(f"Saving {self} to DB")
        session.add(self)
        session.commit()


class ChannelUsers(Base):
    __tablename__ = "channel_users"
    __allow_unmapped__ = True
    # A user can only be in the channel once
    __table_args__ = (UniqueConstraint("user", "channel_id", name="chan_users"),)

    id = sa.Column(sa.Integer, sa.Sequence('users_id_seq'), primary_key=True)
    user = sa.Column(sa.Text, nullable=False)
    channel_id = sa.Column(ForeignKey("channels.id"), nullable=False)
    channel: Mapped[Channel] = relationship(back_populates="users")

    @staticmethod
    def new_user(user: str) -> "ChannelUsers":
        obj = ChannelUsers(user=user)
        return obj

    def __repr__(self):
        return f"<ChannelUser(user='{self.user}', channel='{self.channel}')>"


class ChannelMessages(Base):
    __tablename__ = "channel_messages"
    __allow_unmapped__ = True

    id = sa.Column(sa.Integer, sa.Sequence('message_id_seq'), primary_key=True)
    channel_id = sa.Column(ForeignKey("channels.id"), nullable=False)
    channel: Mapped[Channel] = relationship(back_populates="messages")
    message = sa.Column(sa.Text, nullable=False)
    timestamp = sa.Column(sa.DateTime, server_default=sa.func.now(), nullable=False)

    def __repr__(self):
        return (f"<ChannelMessages(channel='{self.channel.name}', message='{self.message}', "
                f"time='{self.timestamp}')>")

    @staticmethod
    def new_message(packet: aprsd_packets.Packet) -> "ChannelMessages":
        obj = ChannelMessages(message=packet.raw)
        return obj

    @property
    def packet(self):
        pkt = aprsd_packets.Packet.factory(aprslib.parse(self.message))
        pkt.timestamp = self.timestamp
        return pkt



import datetime
import logging
import threading
import signal
import sys
import time

import click
from oslo_config import cfg


import aprsd
from aprsd import cli_helper, client, packets, stats
from aprsd import threads as aprsd_threads
from aprsd.threads import tx, registry, keep_alive
from aprsd.threads import stats as stats_thread

from aprsd_irc_extension.db import models
from aprsd_irc_extension.db import session as db_session
import aprsd_irc_extension
from aprsd_irc_extension import cmds, utils
from aprsd_irc_extension import conf  # noqa


CONF = cfg.CONF
LOG = logging.getLogger("APRSD")


def signal_handler(sig, frame):
    click.echo("signal_handler: called")
    aprsd_threads.APRSDThreadList().stop_all()
    if "subprocess" not in str(frame):
        LOG.info(
            "Ctrl+C, Sending all threads exit! Can take up to 10 seconds {}".format(
                datetime.datetime.now(),
            ),
        )
        time.sleep(1.5)
        packets.PacketTrack().save()
        packets.WatchList().save()
        packets.SeenList().save()
        LOG.info(stats.APRSDStats())
        # signal.signal(signal.SIGTERM, sys.exit(0))
        # sys.exit(0)


class InvalidChannelName(Exception):
    pass


class ChannelService:
    """Class for handling channel related commands."""

    @staticmethod
    def join(channel: str, user: str) -> models.Channel:
        """User wants to join a channel."""
        LOG.info(f"{user} has joined {channel}")
        session = db_session.get_session()
        user_obj = models.ChannelUsers(user=user)
        ch = models.Channel.find_by_name(session, channel)
        ch.users.append(user_obj)
        pkt = packets.MessagePacket(
            from_call=CONF.callsign,
            to_call=user,
            message_text=f"Welcome to channel {channel}",
        )
        tx.send(pkt)
        time.sleep(1)
        tx.send(packets.MessagePacket(
            from_call=CONF.callsign,
            to_call=user,
            message_text=f"Use /leave {channel} to leave",
        ))
        ch.save(session)
        session.remove()
        return channel

    @staticmethod
    def leave(channel: str, user: str) -> models.Channel:
        """User wants to leave a channel."""
        LOG.info(f"{user} wants to leave channel {channel}")
        session = db_session.get_session()
        ch = models.Channel.find_by_name(session, channel)
        LOG.info(repr(ch))
        found = False
        for user_obj in ch.users:
            if user_obj.user == user:
                found = True
                session.delete(user_obj)
                session.flush()
                session.commit()
                pkt = packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=user,
                    message_text=f"Left channel {channel}",
                )
                tx.send(pkt)

        if not found:
            LOG.warning(f"{user} not in channel {channel}")
            pkt = packets.MessagePacket(
                from_call=CONF.callsign,
                to_call=user,
                message_text=f"not in channel {channel}",
            )
            tx.send(pkt)

        session.remove()
        return channel

    @staticmethod
    def add_message(channel: str, pkt: packets.MessagePacket) -> models.Channel:
        """Add a message to the channel."""
        LOG.info(f"Adding message to channel {channel}")
        session = db_session.get_session()
        ch = models.Channel.find_by_name(session, channel)
        ch.messages.append(models.ChannelMessages.new_message(pkt))
        ch.save(session)
        session.remove()
        return channel


class IRChannels:
    """List of IRC Channels."""
    _instance = None
    lock = threading.Lock()
    data: dict = {}
    db_session = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.data = {}
            cls._instance._load()
        return cls._instance

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def _load(self):
        LOG.info("IRChannels: Loading channels from DB")
        session = db_session.get_session()
        channels = session.query(models.Channel).all()
        for ch in channels:
            LOG.info(f"IRChannels: Loading channel {ch}")
            users = []
            for user in ch.users:
                users.append(user.user)
            LOG.info(f"IRChannels: Loading messages {len(ch.messages)}")
            self.data[ch.name] = users

        LOG.info(f"IRChannels: Loaded {len(self.data)} channels from DB")
        session.remove()

    def get(self, id):
        with self.lock:
            return self.data[id]

    def list(self, packet) -> None:
        """Send a list of channels and count of users in channel."""
        channels = models.Channel.get_all_channels()
        user = packet.from_call
        for ch in channels:
            pkt = packets.MessagePacket(
                from_call=CONF.callsign,
                to_call=user,
                message_text=f"Channel {ch.name} Users({len(ch.users)})",
            )
            tx.send(pkt)

    def add_channel(self, name: str):
        if not name.startswith("#"):
            raise InvalidChannelName(
                "Channel name must start with #")
        if name not in self.data:
            models.Channel.create_channel(name)
            # initialize w/o users
            self.data[name] = []

    def remove_channel(self, name: str) -> None:
        if not name.startswith("#"):
            raise InvalidChannelName(
                "Channel name must start with #")
        if name in self.data:
            del self.data[name]

    def get_channel(self, name: str) -> models.Channel:
        if not name.startswith("#"):
            raise InvalidChannelName(
                "Channel name must start with #")
        return self.data.get(name)

    def channel_exists(self, name: str) -> bool:
        if not name.startswith("#"):
            raise InvalidChannelName(
                "Channel name must start with #")
        return name in self.data


class APRSDIRCProcessPacketThread(aprsd_threads.APRSDProcessPacketThread):
    # Commands for IRC Channels
    commands = {
        "/join": {"cmd": "join",
                  "desc": "/join #channel or /j #channel - Join a channel"},
        "/leave": {"cmd": "leave",
                   "desc": "/leave #channel or /l #channel - Leave a channel"},
    }
    short_commands = {
        "/j": {"cmd": "join",
               "desc": "/join #channel or /j #channel - Join a channel"},
        "/l": {"cmd": "leave",
               "desc": "/leave #channel or /l #channel - Leave a channel"},
    }

    server_commands = {
        "/list": {"cmd": "list",
                  "desc": "/list or /ls - list all channels"},
    }
    short_server_commands = {
        "/ls": {"cmd": "list",
                "desc": "/list or /ls - list all channels"},
    }

    def is_channel_command(self, message):
        msg_parts = message.split()
        for command in self.commands:
            if message.startswith(command):
                return True
        for command in self.short_commands:
            if msg_parts[0] == command:
                return True
        return False

    def get_channel_command(self, message):
        msg_parts = message.split()
        for command in self.commands:
            if message.startswith(command):
                return self.commands[command]
        for command in self.short_commands:
            if msg_parts[0] == command:
                return self.short_commands[command]
        return None

    def is_server_command(self, message):
        for command in self.server_commands:
            if message.startswith(command):
                return True
        for command in self.short_server_commands:
            if message.startswith(command):
                return True
        return False

    def get_server_command(self, message):
        for command in self.server_commands:
            if message.startswith(command):
                return self.server_commands[command]
        for command in self.short_server_commands:
            if message.startswith(command):
                return self.short_server_commands[command]
        return None

    def _user_channel_count(self, user):
        """How many channels is a user in?"""
        LOG.info(f"Checking how many channels {user} is in")
        LOG.info(f"IRChannels().data: {IRChannels().data}")
        count = 0
        found = {}
        session = db_session.get_session()
        for name in IRChannels():
            ch = models.Channel.find_by_name(session, name)
            LOG.info(f"Checking channel {ch.name} : {ch.users}")
            for user_ojb in ch.users:
                if user_ojb.user == user:
                    count += 1
                    found[ch.name] = ch.name
                    continue
        session.remove()

        LOG.info(f"User {user} is in {count} channels")
        return count, found

    def process_channel_command(self, packet, command_name, channel_name):
        fromcall = packet.from_call
        message = packet.get("message_text")
        ch = None
        if not channel_name:
            # They didn't specify a channel name
            # If this is leave
            # find how many channels the user is in
            if command_name == "/leave" or command_name == "/l":
                count, found = self._user_channel_count(fromcall)
                if count == 1:
                    channel_name = found.popitem()[1]
                else:
                    channel_names = ", ".join(found.keys())
                    LOG.info(f"User {fromcall} is in {count} channels ({channel_names}). "
                             "Need to specify channel when leaving.")
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text="Need to specify channel when leaving. /leave #channel",
                    ))
                    return

        session = None
        try:
            if IRChannels().channel_exists(channel_name):
                session = db_session.get_session()
                ch = models.Channel.find_by_name(session, channel_name)
        except InvalidChannelName as e:
            LOG.error(f"Failed to add channel: {e}")
            tx.send(packets.MessagePacket(
                from_call=CONF.callsign,
                to_call=fromcall,
                message_text="Channel name must start with #",
            ))
            return
        finally:
            if session:
                session.remove()

        if not ch:
            ch = IRChannels().add_channel(channel_name)

        cmd_dict = self.get_channel_command(message)
        LOG.warning(f"cmd_dict: {cmd_dict}")
        if not cmd_dict:
            LOG.info(f"Unknown command: {command_name}")
            tx.send(
                packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=fromcall,
                    message_text=f"Unknown command: {command_name}",
                )
            )
            return

        svc = ChannelService
        cmd = getattr(svc, cmd_dict["cmd"])
        cmd(channel_name, fromcall)
        return

    def process_irc_command(self, packet):
        message = packet.get("message_text")
        msg_parts = message.split()
        command_name = msg_parts[0]
        try:
            channel_name = msg_parts[1]
        except IndexError:
            channel_name = None
        self.process_channel_command(packet, command_name, channel_name)

    def process_server_command(self, packet):
        fromcall = packet.from_call
        message = packet.get("message_text")
        msg_parts = message.split()
        command_name = msg_parts[0]
        cmd_dict = self.get_server_command(message)
        LOG.warning(f"cmd_dict: {cmd_dict}")
        if not cmd_dict:
            LOG.info(f"Unknown command: {command_name}")
            tx.send(
                packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=fromcall,
                    message_text=f"Unknown command: {command_name}",
                )
            )
            return
        cmd = getattr(IRChannels(), cmd_dict["cmd"])
        cmd(packet)
        return

    def process_our_message_packet(self, packet):
        irc_channels = IRChannels()
        fromcall = packet.from_call
        message = packet.get("message_text")

        # check to see if there are channel commands
        if self.is_channel_command(message):
            LOG.info(f"Processing channel command: {message}")
            self.process_irc_command(packet)
            return
        elif self.is_server_command(message):
            LOG.info(f"Processing server command: {message}")
            self.process_server_command(packet)
            return
        else:
            if message.startswith("help"):
                # They want a list of commands
                LOG.info(f"Send help message to {fromcall}")
                for command in self.commands:
                    cmd = self.commands.get(command)
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text=f"{cmd['desc']}",
                    ))
                for command in self.server_commands:
                    cmd = self.server_commands.get(command)
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text=f"{cmd['desc']}",
                    ))
                return

            # If not a channel command, then it's a message
            # to a channel or user
            channel_name = message.split()[0]
            LOG.info(f"Send message to channel {channel_name}??")
            ch = None
            session = db_session.get_session()
            try:
                ch = irc_channels.get_channel(channel_name)
            except InvalidChannelName as e:
                count, found = self._user_channel_count(fromcall)
                if count == 1:
                    channel_name = found.popitem()[1]
                    message = f"{channel_name} {message}"
                    ch = models.Channel.find_by_name(session, channel_name)
                elif count > 1:
                    LOG.info("Failed to get channel from a user message. User in more than 1 channel")
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text="Need to specify channel when sending a message. '#channel msg'",
                    ))
                    session.remove()
                    return
                elif count == 0:
                    LOG.error(f"User is not in any channels: {e}")
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text="Must join a channel to send a message. try /list",
                    ))
                    session.remove()
                    return
                else:
                    LOG.error(f"Failed to get channel: {e}")
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text="Channel name must start with #",
                    ))
                    session.remove()
                    return

            if ch:
                found = False
                for user_obj in ch.users:
                    if user_obj.user == fromcall:
                        found = True

                if not found:
                    LOG.error(f"{fromcall} not in channel {channel_name}")
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text=f"{fromcall} not in channel {channel_name}",
                    ))
                    tx.send(packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text=f"Send /join {channel_name} to join channel",
                    ))
                    session.remove()
                    return

                msg = message.replace(ch.name, f"{ch.name} {fromcall}")
                for user_obj in ch.users:
                    if user_obj.user != fromcall:
                        tx.send(packets.MessagePacket(
                            from_call=CONF.callsign,
                            to_call=user_obj.user,
                            message_text=msg,
                        ))
                session.remove()
                ChannelService.add_message(ch.name, packet)
            else:
                LOG.error(f"Channel {channel_name} not found")
                tx.send(packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=fromcall,
                    message_text=f"Channel {channel_name} not found",
                ))
                time.sleep(1)
                tx.send(packets.MessagePacket(
                    from_call=CONF.callsign,
                    to_call=fromcall,
                    message_text=f"Use /join {channel_name} to create it",
                ))


class ChannelInfoThread(aprsd_threads.APRSDThread):
    _loop_cnt: int = 1

    def __init__(self):
        super().__init__("ChannelInfo")
        self._loop_cnt = 1

    def loop(self):
        # Only dump out the stats every 60 seconds
        if self._loop_cnt % 60 == 0:
            session = db_session.get_session()
            irc_channels = session.query(models.Channel).all()
            for ch in irc_channels:
                LOG.info(repr(ch))
        self._loop_cnt += 1
        time.sleep(1)
        return True


@cmds.irc.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "-f",
    "--flush",
    "flush",
    is_flag=True,
    show_default=True,
    default=False,
    help="Flush out all old aged messages on disk.",
)
@click.pass_context
@cli_helper.process_standard_options
def server(ctx, flush):
    """Run an APRS IRC server."""
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    level, msg = utils._check_version()
    if level:
        LOG.warning(msg)
    else:
        LOG.info(msg)
    LOG.info(f"APRSD IRC Started version: {aprsd_irc_extension.__version__}")
    LOG.info(f"APRSD version: {aprsd.__version__}")

    # Initialize the client factory and create
    # The correct client object ready for use
    client.ClientFactory.setup()

    # Dump all the config options now.
    CONF.log_opt_values(LOG, logging.DEBUG)

    # Make sure we have 1 client transport enabled
    if not client.factory.is_client_enabled():
        LOG.error("No Clients are enabled in config.")
        sys.exit(-1)

    if not client.factory.is_client_configured():
        LOG.error("APRS client is not properly configured in config file.")
        sys.exit(-1)

    # Now load the msgTrack from disk if any
    packets.PacketList()
    if flush:
        LOG.debug("Deleting saved objects.")
        packets.PacketTrack().flush()
        packets.WatchList().flush()
        packets.SeenList().flush()
    else:
        # Try and load saved MsgTrack list
        LOG.debug("Loading saved objects.")
        packets.PacketTrack().load()
        packets.WatchList().load()
        packets.SeenList().load()

    # Make sure the #lounge channel exists
    IRChannels().add_channel("#lounge")

    stats_store_thread = stats_thread.APRSDStatsStoreThread()
    stats_store_thread.start()

    keepalive = keep_alive.KeepAliveThread()
    keepalive.start()

    rx_thread = aprsd_threads.APRSDDupeRXThread(
        packet_queue=aprsd_threads.packet_queue,
    )
    process_thread = APRSDIRCProcessPacketThread(
        packet_queue=aprsd_threads.packet_queue,
    )
    channel_info_thread = ChannelInfoThread()

    if CONF.aprs_registry.enabled:
        LOG.info("Registry Enabled.  Starting Registry thread.")
        registry_thread = registry.APRSRegistryThread()
        registry_thread.start()

    if CONF.enable_beacon:
        LOG.info("Beacon Enabled.  Starting Beacon thread.")
        bcn_thread = tx.BeaconSendThread()
        bcn_thread.start()

    rx_thread.start()
    process_thread.start()
    channel_info_thread.start()
    packets.PacketTrack()

    rx_thread.join()
    keepalive.join()



"""empty message

Revision ID: 030cab1951e1
Revises:
Create Date: 2024-03-09 15:19:47.143907

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '030cab1951e1'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'channels',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'channel_messages',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('channel_id', sa.Integer(), nullable=False),
        sa.Column('message', sa.Text(), nullable=False),
        sa.Column('timestamp', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
        sa.ForeignKeyConstraint(['channel_id'], ['channels.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'channel_users',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user', sa.Text(), nullable=False),
        sa.Column('channel_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['channel_id'], ['channels.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('user', 'channel_id', name='chan_users')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('channel_users')
    op.drop_table('channel_messages')
    op.drop_table('channels')
    # ### end Alembic commands ###



from oslo_config import cfg


irc_group = cfg.OptGroup(
    name="aprsd_irc_extension",
    title="APRSD IRC Extension settings",
)

irc_opts = [
    cfg.BoolOpt(
        "enabled",
        default=False,
        help="Enable the plugin?",
    ),
    cfg.StrOpt(
        "default_channel",
        default="#lounge",
        help="This default channel will always exist.",
    ),
    cfg.StrOpt(
        "db_dsn",
        default="sqlite:////tmp/aprsd-irc.db",
        help="The DSN URI for the database",
    ),
]

ALL_OPTS = irc_opts


def register_opts(cfg):
    cfg.register_group(irc_group)
    cfg.register_opts(ALL_OPTS, group=irc_group)


def list_opts():
    return {
        irc_group.name: irc_opts,
    }
